# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

proteinseq_toks = {
    'toks': ['L', 'A', 'G', 'V', 'S', 'E', 'R', 'T', 'I', 'D', 'P', 'K', 'Q', 'N', 'F', 'Y', 'M', 'H', 'W', 'C', 'X', 'B', 'U', 'Z', 'O', '.', '-']
}

rnaseq_toks = {
    'toks': ['A', 'C', 'G', 'U', 'R', 'Y', 'K', 'M', 'S', 'W', 'B', 'D', 'H', 'V', 'N', '-']
}

rnaseq_3mer_toks = {
    'toks': [
        "GAG", "AAG", "GAA", "CUG", "CAG", "GAU", "AAA", "GUG",
        "GAC", "AUG", "GCC", "AAC", "GCU", "AAU", "AUC", "UUC",
        "GGA", "AUU", "GGC", "UUU", "CCA", "AGC", "GCA", "UCU",
        "CUC", "ACC", "CAA", "CCU", "UCC", "ACA", "UUG", "GUU",
        "CUU", "UAC", "ACU", "CCC", "UCA", "GUC", "GGU", "CAC",
        "AGU", "UAU", "AGA", "CAU", "GGG", "UGG", "UGC", "AGG",
        "UGU", "AUA", "CGC", "UUA", "GCG", "CGG", "CCG", "GUA",
        "CUA", "ACG", "UCG", "CGA", "CGU", "UGA", "UAA", "UAG",
    ]
}